#!/usr/bin/python3

###########################################################
#
#
# Script permettant de recréer une table idmap propre.
# Todo : Corriger les well known sid avec un id fixe.
# De base, offset à 10 000.
# Offset system modifie à 1 000 000 (pour les well know sid
# YA.
#
###########################################################

VERSION="1.0"

# -*- coding: utf-8 -*-
import shutil
import os
from samba.auth import system_session
from samba.credentials import Credentials
from samba.param import LoadParm
from samba.samdb import SamDB
import sys
import configparser
import argparse
import subprocess
#import mmh3


#############
# Variables #
#############
path_idmap = "/var/lib/samba/private/idmap.ldb"
path_idmap_dryrun = "/root/idmap.ldb"
path_samba_private = "/var/lib/samba/private/sam.ldb"



def sid_to_str(sid):
    try:
        # Python 3
        if str is not bytes:
            # revision
            revision = int(sid[0])
            # count of sub authorities
            sub_authorities = int(sid[1])
            # big endian
            identifier_authority = int.from_bytes(sid[2:8], byteorder='big')
            # If true then it is represented in hex
            if identifier_authority >= 2 ** 32:
                identifier_authority = hex(identifier_authority)

            # loop over the count of small endians
            sub_authority = '-' + '-'.join([str(int.from_bytes(sid[8 + (i * 4): 12 + (i * 4)], byteorder='little')) for i in range(sub_authorities)])
        # Python 2
        else:
            revision = int(b2a_hex(sid[0]))
            sub_authorities = int(b2a_hex(sid[1]))
            identifier_authority = int(b2a_hex(sid[2:8]), 16)
            if identifier_authority >= 2 ** 32:
                identifier_authority = hex(identifier_authority)

            sub_authority = '-' + '-'.join([str(int(b2a_hex(sid[11 + (i * 4): 7 + (i * 4): -1]), 16)) for i in range(sub_authorities)])
        objectSid = 'S-' + str(revision) + '-' + str(identifier_authority) + sub_authority

        return objectSid
    except Exception:
        pass

    return sid


def get_sssd_offset(sid):
    # https://pagure.io/SSSD/sssd/blob/master/f/src/lib/idmap/sss_idmap_private.h
    IDMAP_UPPER = 2000200000
    IDMAP_LOWER = 200000
    IDMAP_RANGE = 200000

    max_slices = (IDMAP_UPPER - IDMAP_LOWER) / IDMAP_RANGE

    splitsid = sid.rpartition('-')
    domain = splitsid[0]
    rid = int(splitsid[2])

    # Generate MurmurHash3 of Domain SID
    #domain_hash = mmh3.hash(domain, 0xdeadbeef, signed=False)

    # Generate slice
    slice = domain_hash % max_slices

    # Calculate offset
    offset = (slice * IDMAP_RANGE) + IDMAP_LOWER

    # Print some info
    return offset


##########
# PARSER #
##########
parser = argparse.ArgumentParser(description='Python script to recreate idmap.ldb')
parser.add_argument("--execute", 
  dest="execute", 
  action="store_true", 
  help="Launch script on production mode. Please use on dry-run first")
# Exclusion
req = parser.add_mutually_exclusive_group(required=True)
req.add_argument("--use-sssd-offset",
  dest="sssd-offset",
  action="store_true",
  help="SSSD automatic offset. Need to have sss-calculate-hash.py. Can be downloaded here : https://gist.github.com/jfautley/4643ab8f6078253a87a2c71489bb85ee")
req.add_argument("--offset",
  dest="offset",
  action="store",
  help="Offset for idmap range. Default : 10.000")
args = parser.parse_args()
options = vars(args)


if not options["execute"]:
    print("We work on a copy of idmap.ldb.")
    # On copie le fichier idmap.ldb de production dans /root/. (pas de travail en prod)
    shutil.copy2(path_idmap, path_idmap_dryrun)
    path_idmap = path_idmap_dryrun


# Connexion à samdb
lp = LoadParm()
creds = Credentials()
creds.guess(lp)


# On request tous les objectsid du domaine
samdb_samba = SamDB(url=path_samba_private, session_info=system_session(),credentials=creds, lp=lp)
samdb_idmap = SamDB(url=path_idmap, session_info=system_session(),credentials=creds, lp=lp)

result_samba = samdb_samba.search(expression="(&(objectSid=*)(|(objectclass=user)(objectclass=computer)(objectclass=group)(objectclass=foreignSecurityPrincipal)(objectclass=builtinDomain)))",attrs=["dn","objectSid"],controls=["search_options:1:2"])
result_idmap = samdb_idmap.search(expression="xidNumber=*",attrs=["dn","xidNumber"])

dict_wellknown = {
    "S-1-1-0":2,
    "S-1-3-0":3,
    "S-1-3-1":4,
    "S-1-3-4":5,
    "S-1-5-1":6,
    "S-1-5-10":7,
    "S-1-5-1000":8,
    "S-1-5-11":9,
    "S-1-5-12":10,
    "S-1-5-13":11,
    "S-1-5-14":12,
    "S-1-5-15":13,
    "S-1-5-17":14,
    "S-1-5-18":15,
    "S-1-5-19":16,
    "S-1-5-2":17,
    "S-1-5-20":18,
    "S-1-5-3":19,
    "S-1-5-32":20,
    "S-1-5-32-544":21,
    "S-1-5-32-545":22,
    "S-1-5-32-546":23,
    "S-1-5-32-548":24,
    "S-1-5-32-549":25,
    "S-1-5-32-550":26,
    "S-1-5-32-551":27,
    "S-1-5-32-552":28,
    "S-1-5-32-554":29,
    "S-1-5-32-555":30,
    "S-1-5-32-556":31,
    "S-1-5-32-557":32,
    "S-1-5-32-558":33,
    "S-1-5-32-559":34,
    "S-1-5-32-560":35,
    "S-1-5-32-561":36,
    "S-1-5-32-562":37,
    "S-1-5-32-568":38,
    "S-1-5-32-569":39,
    "S-1-5-32-573":40,
    "S-1-5-32-574":41,
    "S-1-5-4":42,
    "S-1-5-6":43,
    "S-1-5-64-10":44,
    "S-1-5-64-14":45,
    "S-1-5-64-21":46,
    "S-1-5-7":47,
    "S-1-5-8":48,
    "S-1-5-9":49,
                }


## Si sssd-offset selectionne, alors on essaye de calculer l'offset depuis le fichier sss-calculate-hash.py du repertoire courant
if options["sssd-offset"] is True:
    print("Try to calculate offset like sssd with sss-calculate-hash.py ...")
    # on recupere une entree
    for entry in result_samba:
        if "S-1-5-21-" in sid_to_str(entry['objectSid'][0]):
            usersid=sid_to_str(entry['objectSid'][0])
            break
    offset = int(str(get_sssd_offset(usersid)).split('.')[0])
    offset_system = offset + 1000000
    print(str(offset) + " " + str(offset_system))
else:
    offset = int(options["offset"])
    print("Use offset of %s" %offset)
    offset_system = offset + 1000000

def generate_ldif(fdn=None,frid=None):
    global ldif_data
    # Purge de la variable ldif_data
    ldif_data = ""
    # Check si dn existe et rid est numérique
    if fdn == None or str(frid).isnumeric() != "True":
#        ldif_data = """dn: %s
#changetype: modify
#replace: xidNumber
#xidNumber: %s
#""" % (fdn,frid)
        ldif_data = """dn: CN=%s
changetype: add
objectClass: sidMap
objectSid: %s
type: ID_TYPE_BOTH
xidNumber: %s
distinguishedName: CN=%s
cn: %s
""" % (fdn,fdn,frid,fdn,fdn)
        return ldif_data
    else:
        print("Erreur avec le ldif généré.")
        sys.exit(1)

def delete_entry(sid):
   try:
       samdb_idmap.delete("CN=%s" % sid)
   except:
       print("CN=%s not registered" % sid)

# Gestion de l'entrée config

xidnumber_config = str(offset) 
lowerbound_config = str(offset)
upperbound_config = str(int(offset + 1000000))

ldif_config = """dn: CN=CONFIG
changetype: add
cn: CONFIG
xidNumber: %s
distinguishedName: CN=CONFIG
lowerBound: %s
upperBound: %s
""" %(xidnumber_config, lowerbound_config, upperbound_config)

configactual =  samdb_idmap.search(expression="cn=config")[0]
if str(configactual["xidNumber"]) != xidnumber_config or str(configactual["lowerBound"]) != lowerbound_config or str(configactual["upperBound"]) != upperbound_config:
    delete_entry('CONFIG')
    samdb_idmap.modify_ldif(ldif_config)

# Gestion des dict

dict_sid_samba = {}
dict_sid_idmap = {}

#Création du dict samba
print("Starting remapping")
for entry in result_samba:
    sid = sid_to_str(entry['objectSid'][0])
    if sid in dict_wellknown:
        dict_sid_samba[sid] = offset_system + dict_wellknown[sid]
        #print(str(dict_sid_samba[sid]))
        continue
    dict_sid_samba[sid] = int(str(sid).split('-')[7]) + offset

#Création du dict_idmap
for entry in result_idmap:
    dict_sid_idmap[str(entry['dn']).split('=',1)[1]] = int(str(entry["xidNumber"]))

restart_need = False

#Comparaison des entrée présente dans les deux
for entry in dict_sid_idmap:
    if entry in dict_sid_samba:
        if dict_sid_idmap[entry] != dict_sid_samba[entry]:
            delete_entry(entry)
            ldif_data = generate_ldif(entry, dict_sid_samba[entry])
            print(ldif_data)
            samdb_idmap.modify_ldif(ldif_data)
            restart_need = True

#Création des entrée manquante
for entry in dict_sid_samba:
    if not entry in dict_sid_idmap:
        ldif_data = generate_ldif(entry, dict_sid_samba[entry])
        samdb_idmap.modify_ldif(ldif_data)
        restart_need = True

# Delete missing entry
for entry in dict_sid_idmap:
    if not entry in dict_sid_samba:
        if not 'S-1-5-21-' in entry:
            continue
        print('Delete entry %s' % entry)
        delete_entry(entry)

if restart_need:
    print("Done. Please verify the modified file with this command :")
    print("  ldbsearch -H %s xidNumber=* | grep -e ^dn -e xidNumber " % (path_idmap))
    print("")
    print("If launched in dry-run mode, please do : cp -f %s %s", path_idmap_dryrun, path_idmap)
    print("")
    print("And if all is good, do :")
    print("On debian :")
    print("  systemctl stop samba-ad-dc; net cache flush; systemctl start samba-ad-dc; samba-tool ntacl sysvolreset")
    print("")
    print("On Centos / redhat")
    print("  systemctl stop samba; net cache flush; systemctl start samba; samba-tool ntacl sysvolreset")


